/*
 * heap.c
 *
 * C code file for an implementation of a heap to be used by the accompanying
 * Matlab lattice decoders. Can be compiled into a standalone library or using
 * mex into a Matlab library. Pre-compiled for Matlab 7.0.1 R14SP1 (Linux) and
 * Matlab 6.5 R13 (Windows); for best effect, re-compile for other releases.
 *
 * Version 1.0, copyright 2006 by Karen Su (karen.su@utoronto.ca).
 * Version 1.1
 *   Fixed heapDelMin; added heapGetMaxSize, heapGetDim and appropriate interfaces.
 * Version 1.2
 *   Fixed heapInsert; added heap.pLastDel to replace global pLastDelNode;
 *   modified heapGetMin to return key == -1 (invalid value) if heap is empty.
 */

#include "stdlib.h"
#include "stdio.h"
#include "malloc.h"
#include "memory.h"
#include "heap.h"
#include "math.h"
#include "string.h"

/* #define DEBUG */
/* #define VER65 */
#define VER7R14
#define MAXNODES 32768
#define COMPILEFORMATLAB
/* #define MSVC */

#ifdef VER65
#define mycreatedoublescalar mxCreateDoubleScalar
#define mystrncasecmp strncasecmp
#else
#define mycreatedoublescalar mxCreateScalarDouble
#define mystrncasecmp strncasecmp
#endif

#ifdef VER7R14
#define myprintf printf
#define myerrmsg(str) { printf(str); return; }
#else
#define myprintf mexPrintf
#define myerrmsg mexErrMsgTxt
#endif

#ifdef MSVC
#define strncasecmp strnicmp
#endif

#ifdef COMPILEFORMATLAB
#include "mex.h"
#define mymalloc mxCalloc
#define myfree mxFree
#else
#define mymalloc calloc
#define myfree free
#endif

struct heap *heapInit(int spdim, int type) {
  return heapInitT(spdim, MAXNODES, type);
}

struct heap *heapInitT(int spdim, int T, int type) {
  struct heap *pH = (struct heap*)mymalloc(1, sizeof(struct heap));
  pH->nNodes   = 0;
  pH->pRoot    = NULL;
  pH->spdim    = spdim;
  pH->maxNodes = T;
  pH->type     = type;
  pH->pLastDel = NULL;

#ifdef COMPILEFORMATLAB
  /* Ensure that matlab doesn't erase this */
  mexMakeMemoryPersistent((void*)pH);
#endif

  return pH;
}

struct hNode *heapNewNode(double       key,
                          double      *pY,
                          double      *pZ,
                          int          dim,
                          struct heap *pH) {
  struct hNode *pNewNode;

  if (pH->pLastDel) {
    pNewNode     = pH->pLastDel;
    pH->pLastDel = NULL;
  } else {
    switch (pH->type) {
      case BASIC:
        pNewNode = (struct hNode *)mymalloc(1, sizeof(struct hNode));
        break;
      case LATTICE:
        pNewNode = (struct hNode *)mymalloc(1, sizeof(struct hNodeI));
        break;
      case LAST:
        pNewNode = (struct hNode *)mymalloc(1, sizeof(struct hNodeL));
        break;
      default:
        break;
    }

    if (pNewNode) {  /* mxCalloc didn't fail */
#ifdef COMPILEFORMATLAB
      mexMakeMemoryPersistent((void *)pNewNode);
#endif
    } else {
      return NULL;
    }

    pNewNode->pY      = (double *)mymalloc(pH->spdim, sizeof(double));

    if (pNewNode->pY) {  /* mxCalloc didn't fail */
#ifdef COMPILEFORMATLAB
      mexMakeMemoryPersistent((void *)pNewNode->pY);
#endif
    } else {
      myfree(pNewNode);
      return NULL;
    }
  }

  pNewNode->key     = key;
  if (pY) { memcpy(pNewNode->pY, pY, dim*sizeof(double)); }
  pNewNode->pZ      = pNewNode->pY+dim;
  if (pZ) { memcpy(pNewNode->pZ, pZ, (pH->spdim-dim)*sizeof(double)); }
  pNewNode->dim     = dim;

  pNewNode->pLCh    = NULL;
  pNewNode->pRCh    = NULL;

  return pNewNode;
}

int heapInsertB(double       key, 
                double      *pY, 
                double      *pZ, 
                int          dim, 
                struct heap *pH) {
  struct hNode *pNewNode = heapNewNode(key, pY, pZ, dim, pH);

  if (pNewNode) { 
    return heapInsert(pNewNode, pH);
  } else {
    return -1;
  }
}

int heapInsertI(double       key, 
                double      *pY, 
                double      *pZ, 
                int          dim, 
                int          cnum, 
                double       pweight, 
                double       utarget, 
                double       delta, 
                struct heap *pH) {
  struct hNodeI *pNewNode = (struct hNodeI *)heapNewNode(key, pY, pZ, dim, pH);

  if (pNewNode) {
    pNewNode->cnum    = cnum;
    pNewNode->pweight = pweight;
    pNewNode->utarget = utarget;
    pNewNode->delta   = delta;

    return heapInsert((struct hNode *)pNewNode, pH);
  } else {
    return -1;
  }
}

int heapInsertL(double       key, 
                double      *pY, 
                double      *pZ, 
                int          dim, 
                int          cnum, 
                double       pweight, 
                double       utarget, 
                double       delta, 
                double      *pA,
                double       sqradius,
                int          lbound,
                int          ubound,
                double       uoffset,
                struct heap *pH) {
  struct hNodeL *pNewNode = (struct hNodeL *)heapNewNode(key, pY, pZ, dim, pH);

  if (pNewNode) {
    pNewNode->cnum    = cnum;
    pNewNode->pweight = pweight;
    pNewNode->utarget = utarget;
    pNewNode->delta   = delta;

    if (dim>0) {
      pNewNode->pA = (double *)mymalloc(dim, sizeof(double));

      if (pNewNode->pA) {  /* mxCalloc didn't fail */
#ifdef COMPILEFORMATLAB
        mexMakeMemoryPersistent((void *)pNewNode->pA);
#endif
      } else {
        hNodeFree((struct hNode *)pNewNode);
        return -1;
      }
      if (pA) { memcpy(pNewNode->pA, pA, dim*sizeof(double)); }
    }
    pNewNode->sqradius = sqradius;
    pNewNode->lbound   = lbound;
    pNewNode->ubound   = ubound;
    pNewNode->uoffset  = uoffset;
  
    return heapInsert((struct hNode *)pNewNode, pH);
  } else {
    return -1;
  }
}

int heapInsert(struct hNode *pNewNode, struct heap *pH) {
  struct hNode  *pParentNode, *pLastParNode;
  int            nextNodeIndex, ii, nSwaps = 0;

  pH->nNodes++;
  nextNodeIndex = pH->nNodes;

#ifdef DEBUG
  printf("InsertIng node %i: ",nextNodeIndex); 
#endif

  if (pH->pRoot) { /* find where to append new node, bubble */
    pParentNode = pH->pRoot;

    for (ii = (int)floor(log(nextNodeIndex)/log(2)-1); ii > 0; ii--) {
      if (nextNodeIndex&(1<<ii)) {       /* follow right child path */

#ifdef DEBUG
        printf("Right "); 
#endif 

        pParentNode = pParentNode->pRCh; 
      } else {                           /* follow left child path */

#ifdef DEBUG
        printf("Left "); 
#endif 

        pParentNode = pParentNode->pLCh; 
      }
    }
    if (nextNodeIndex&1) {           /* append new node as right child */

#ifdef DEBUG
      printf("InsertasRight\n"); 
#endif

      pParentNode->pRCh = pNewNode; 
    } else {                         /* append new node as left child */

#ifdef DEBUG
      printf("InsertasLeft\n"); 
#endif

      pParentNode->pLCh = pNewNode; 
    }
    pNewNode->pPar = pParentNode; 

#ifdef DEBUG
    printf("Intermediate heap:\n");
    heapPrintContents(pH);
#endif

    pLastParNode = pNewNode->pPar;      /* this node may be it... */
    nSwaps = heapBubbleUp(pH, pNewNode); 

    if (pH->nNodes > pH->maxNodes) {    /* delete last node, can only save T */
                                        /* note that this is not the largest */
                                        /* weight node. */

      if (pNewNode->pPar != pLastParNode) {  /* pLastParNode was bubbled last*/
        if (pLastParNode->pPar->pLCh == pLastParNode) {
          pLastParNode->pPar->pLCh = NULL;
        } else if (pLastParNode->pPar->pRCh == pLastParNode) {
          pLastParNode->pPar->pRCh = NULL;
        }
        pH->pLastDel = pLastParNode; /* we do not delete because this fragments */
                                     /* memory, instead we will reuse the space */

      } else {                       /* pNewNode not bubbled and is still last  */
        if (pLastParNode->pLCh == pNewNode) {
          pLastParNode->pLCh = NULL;
        } else if (pLastParNode->pRCh == pNewNode) {
          pLastParNode->pRCh = NULL;
        }
        pH->pLastDel = pNewNode; 
      }
      pH->nNodes--;
    }
    
  } else { 

#ifdef DEBUG
    printf("InsertasRoot\n"); 
#endif

    pH->pRoot      = pNewNode;
    pNewNode->pPar = NULL;
  } 

  return nSwaps;
}

struct hNode *heapGetMin(struct heap *pH) {
  return pH->pRoot;
}

int heapReplaceMinb(double       key, 
                    double      *pY, 
                    double      *pZ, 
                    int          dim, 
                    struct heap *pH) {
  struct hNode *pMinNode;
  int           nSwaps = 0;

  pMinNode = heapGetMin(pH);
  if (pMinNode) {                      
    pMinNode->key     = key;
    if (pY) { memcpy(pMinNode->pY, pY, dim*sizeof(double)); }
    pMinNode->pZ      = pMinNode->pY+dim;
    if (pZ) { memcpy(pMinNode->pZ, pZ, (pH->spdim-dim)*sizeof(double)); }
    pMinNode->dim     = dim;

#ifdef DEBUG
    printf("Intermediate heap:\n");
    heapPrintContents(pH);
#endif

    if (pH->pRoot->pLCh) { nSwaps = heapBubbleDown(pH); }

  } else { nSwaps = heapInsertB(key, pY, pZ, dim, pH); }
  
  return nSwaps;
}

int heapReplaceMini(double       key, 
                    double      *pY, 
                    double      *pZ, 
                    int          dim, 
                    int          cnum, 
                    double       pweight, 
                    double       utarget, 
                    double       delta, 
                    struct heap *pH) {
  struct hNodeI *pMinNodei;
  int            nSwaps = 0;

  pMinNodei = (struct hNodeI *)heapGetMin(pH);
  if (pMinNodei) {                      
    pMinNodei->key     = key;
    if (pY) { memcpy(pMinNodei->pY, pY, dim*sizeof(double)); }
    pMinNodei->pZ      = pMinNodei->pY+dim;
    if (pZ) { memcpy(pMinNodei->pZ, pZ, (pH->spdim-dim)*sizeof(double)); }
    pMinNodei->dim     = dim;

    pMinNodei->cnum    = cnum;
    pMinNodei->pweight = pweight;
    pMinNodei->utarget = utarget;
    pMinNodei->delta   = delta;

#ifdef DEBUG
    printf("Intermediate heap:\n");
    heapPrintContents(pH);
#endif

    if (pH->pRoot->pLCh) { nSwaps = heapBubbleDown(pH); }

  } else { nSwaps = heapInsertI(key, pY, pZ, dim, cnum, pweight, utarget, delta, pH); }
  
  return nSwaps;
}

int heapReplaceMinl(double       key, 
                    double      *pY, 
                    double      *pZ, 
                    int          dim, 
                    int          cnum, 
                    double       pweight, 
                    double       utarget, 
                    double       delta, 
                    double      *pA,
                    double       sqradius,
                    int          lbound,
                    int          ubound,
                    double       uoffset,
                    struct heap *pH) {
  struct hNodeL *pMinNodel;
  int            nSwaps = 0;

  pMinNodel = (struct hNodeL *)heapGetMin(pH);
  if (pMinNodel) {                      
    pMinNodel->key     = key;
    if (pY) { memcpy(pMinNodel->pY, pY, dim*sizeof(double)); }
    pMinNodel->pZ      = pMinNodel->pY+dim;
    if (pZ) { memcpy(pMinNodel->pZ, pZ, (pH->spdim-dim)*sizeof(double)); }
    pMinNodel->dim     = dim;

    pMinNodel->cnum    = cnum;
    pMinNodel->pweight = pweight;
    pMinNodel->utarget = utarget;
    pMinNodel->delta   = delta;

    if (pA) { memcpy(pMinNodel->pA, pA, dim*sizeof(double)); }
    pMinNodel->sqradius = sqradius;
    pMinNodel->lbound   = lbound;
    pMinNodel->ubound   = ubound;
    pMinNodel->uoffset  = uoffset;

#ifdef DEBUG
    printf("Intermediate heap:\n");
    heapPrintContents(pH);
#endif

    if (pH->pRoot->pLCh) { nSwaps = heapBubbleDown(pH); }

  } else { nSwaps = heapInsertI(key, pY, pZ, dim, cnum, pweight, utarget, delta, pH); }
  
  return nSwaps;
}

int heapDelMini(struct heap *pH) {
  struct hNode *pNode  = heapGetMin(pH);
  int           nSwaps = heapDelMin(pH);

  if (pNode) {
    pNode->pPar = NULL;
    pNode->pLCh = NULL;
    pNode->pRCh = NULL;
    hNodeFree(pNode);
  }

  return nSwaps;
}

int heapDelMinl(struct heap *pH) {
  struct hNode *pNode  = heapGetMin(pH);
  int           nSwaps = heapDelMin(pH);

  if (pNode) {
    pNode->pPar = NULL;
    pNode->pLCh = NULL;
    pNode->pRCh = NULL;
    hNodeLFree((struct hNodeL *)pNode);
  }

  return nSwaps;
}

int heapDelMin(struct heap *pH) {
  struct hNode *pDelNode, *pParentNode;
  int           lastNodeIndex, ii, nSwaps = 0;

  pDelNode      = heapGetMin(pH);
  lastNodeIndex = pH->nNodes;

#ifdef DEBUG
  printf("Promoting node %i: ",lastNodeIndex);
#endif

  if (pDelNode) {         /* there is at least one node in the heap */

    if (pDelNode->pLCh) { /* find which node to promote, bubble */
      pParentNode = pDelNode;

      for (ii = (int)floor(log(lastNodeIndex)/log(2)-1); ii > 0; ii--) {
        if (lastNodeIndex&(1<<ii)) {        /* follow right child path */

#ifdef DEBUG
          printf("Right "); 
#endif

          pParentNode = pParentNode->pRCh; 
        } else {                           /* follow left child path */

#ifdef DEBUG
          printf("Left "); 
#endif
        
          pParentNode = pParentNode->pLCh; 
        }
      }

      if (lastNodeIndex&1) {           /* promote right child */

#ifdef DEBUG
        printf("PromoteRight\n"); 
#endif
    
        pH->pRoot       = pParentNode->pRCh;
        pH->pRoot->pPar = NULL;

        /* link left child */
        pH->pRoot->pLCh = pDelNode->pLCh;
        if (pDelNode->pLCh) { pDelNode->pLCh->pPar = pH->pRoot; } 

        /* link right child if not circular */
        if (pDelNode->pRCh != pH->pRoot) {
          pH->pRoot->pRCh = pDelNode->pRCh;
          if (pDelNode->pRCh) { pDelNode->pRCh->pPar = pH->pRoot; }
        }

        /* unlink old parent */
        pParentNode->pRCh = NULL;

      } else {                         /* promote left child */

#ifdef DEBUG
        printf("PromoteLeft\n"); 
#endif
   
        pH->pRoot       = pParentNode->pLCh;
        pH->pRoot->pPar = NULL;

        /* link right child */
        pH->pRoot->pRCh = pDelNode->pRCh;
        if (pDelNode->pRCh) { pDelNode->pRCh->pPar = pH->pRoot; }

        /* link left child if not circular */
        if (pDelNode->pLCh != pH->pRoot) {
          pH->pRoot->pLCh = pDelNode->pLCh;
          if (pDelNode->pLCh) { pDelNode->pLCh->pPar = pH->pRoot; }
        }

        /* unlink old parent */
        pParentNode->pLCh = NULL;
      } 

#ifdef DEBUG
      printf("Intermediate heap:\n");
      heapPrintContents(pH);
#endif

      nSwaps = heapBubbleDown(pH);
    } else {             /* pDelNode != NULL, but has no children */

#ifdef DEBUG
      printf("NoNodestoPromote\n"); 
#endif

      pH->pRoot = NULL;
    }
  
    pH->nNodes--; 

  } else {               /* pDelNode == NULL */

#ifdef DEBUG
    printf("NoNodesinHeap\n"); 
#endif

  }

  return nSwaps;
}

int heapBubbleUp(struct heap *pH, struct hNode *pNewNode) {
  struct hNode *pTempPar, *pTempLCh, *pTempRCh;
  int           nSwaps = 0;

#ifdef DEBUG
  printf("BubblingUp: ");
#endif

  while (pNewNode->pPar && ( pNewNode->key < pNewNode->pPar->key ||
			     ( pNewNode->key == pNewNode->pPar->key &&
			       pNewNode->dim < pNewNode->pPar->dim) ) ) {
    pTempPar = pNewNode->pPar;
    pTempRCh = pNewNode->pRCh;
    pTempLCh = pNewNode->pLCh;
    nSwaps++;

    /* link grandparent or root as parent */
    if (pTempPar->pPar) { 
      if (pTempPar->pPar->pLCh == pTempPar) { /* TempPar was left child */
        pTempPar->pPar->pLCh = pNewNode;
      } else {                                /* TempPar was right child */
        pTempPar->pPar->pRCh = pNewNode;
      }
      pNewNode->pPar = pTempPar->pPar;
    } else {
      pH->pRoot = pNewNode;
      pNewNode->pPar = NULL;
    }

    if (pTempPar->pLCh == pNewNode) {  /* parent becomes left child */

      /* link right sibling as right child */
      pNewNode->pRCh = pTempPar->pRCh;
      if (pTempPar->pRCh) { pTempPar->pRCh->pPar = pNewNode; }

      /* link parent as left child */
      pNewNode->pLCh = pTempPar;
      pTempPar->pPar = pNewNode;

#ifdef DEBUG
      printf("SwapLeft "); 
#endif

    } else {                                 /* parent becomes right child */

      /* link left sibling as left child */
      pNewNode->pLCh = pTempPar->pLCh;
      if (pTempPar->pLCh) { pTempPar->pLCh->pPar = pNewNode; }

      /* link parent as right child */
      pNewNode->pRCh = pTempPar;
      pTempPar->pPar = pNewNode;

#ifdef DEBUG
      printf("SwapRight "); 
#endif

    }

    /* link left child as left grandchild */
    pTempPar->pLCh = pTempLCh;
    if (pTempPar->pLCh) { pTempPar->pLCh->pPar = pTempPar; }

    /* link right child as right grandchild */
    pTempPar->pRCh = pTempRCh;
    if (pTempPar->pRCh) { pTempPar->pRCh->pPar = pTempPar; }

  }  /* while */

#ifdef DEBUG
  printf("\n"); 
#endif

  return nSwaps;
}

int heapBubbleDown(struct heap *pH) {
  struct hNode *pTempNode, *pTempLCh, *pTempRCh, *pNewNode;
  bool          swapLCh, swapRCh;
  int           nSwaps = 0;

#ifdef DEBUG
  printf("BubblingDown: ");
#endif

  /* always bubble down from the root */
  pNewNode = pH->pRoot;
  swapLCh = pNewNode->pLCh && 
	      ( pNewNode->pLCh->key < pNewNode->key ||
	        ( pNewNode->pLCh->key == pNewNode->key && 
	          pNewNode->pLCh->dim < pNewNode->dim)  );
  swapRCh = pNewNode->pRCh &&
	      ( pNewNode->pRCh->key < pNewNode->key ||
	        ( pNewNode->pRCh->key == pNewNode->key && 
	          pNewNode->pRCh->dim < pNewNode->dim)  );
  if (swapLCh && swapRCh) {
    swapLCh = ( pNewNode->pLCh->key < pNewNode->pRCh->key ||
	        ( pNewNode->pLCh->key == pNewNode->pRCh->key && 
	          pNewNode->pLCh->dim <= pNewNode->pRCh->dim)  );
    swapRCh = !swapLCh;
  }

  while (swapLCh || swapRCh) {
    nSwaps++;

    if (swapLCh) {                           /* left child becomes parent */
      pTempNode = pNewNode->pLCh;
      pTempLCh  = pTempNode->pLCh;
      pTempRCh  = pTempNode->pRCh;

      /* link right child as right sibling */
      pTempNode->pRCh = pNewNode->pRCh;
      if (pNewNode->pRCh) { pNewNode->pRCh->pPar = pTempNode; }

      /* link left child as parent */
      pTempNode->pLCh = pNewNode;
      /* Can't do this here cuz still need pNewNode->pPar */
      /* pNewNode->pPar = pTempNode; */

#ifdef DEBUG
      printf("SwapLeft "); 
#endif

    } else {                                 /* right child becomes parent */
      pTempNode = pNewNode->pRCh;
      pTempLCh  = pTempNode->pLCh;
      pTempRCh  = pTempNode->pRCh;

      /* link left child as left sibling */
      pTempNode->pLCh = pNewNode->pLCh;
      if (pNewNode->pLCh) { pNewNode->pLCh->pPar = pTempNode; }

      /* link right child as parent */
      pTempNode->pRCh = pNewNode;
      /* Can't do this here cuz still need pNewNode->pPar */
      /* pNewNode->pPar = pTempNode; */

#ifdef DEBUG
      printf("SwapRight "); 
#endif

    }

    /* link appropriate child (TempNode) to parent or root */
    if (pNewNode->pPar) { 
      if (pNewNode->pPar->pLCh == pNewNode) { /* NewNode was left child */
        pNewNode->pPar->pLCh = pTempNode;
      } else {                                /* NewNode was right child */
        pNewNode->pPar->pRCh = pTempNode;
      }
      pTempNode->pPar = pNewNode->pPar;
    } else {
      pH->pRoot = pTempNode;
      pTempNode->pPar = NULL;
    }

    /* Finish linking appropriate child as parent */
    pNewNode->pPar = pTempNode; 

    /* link left grandchild as left child */
    pNewNode->pLCh = pTempLCh;
    if (pTempLCh) { pTempLCh->pPar = pNewNode; }

    /* link right grandchild as right child */
    pNewNode->pRCh = pTempRCh;
    if (pTempRCh) { pTempRCh->pPar = pNewNode; }

    /* recompute loop condition */
    swapLCh = pNewNode->pLCh && 
	        ( pNewNode->pLCh->key < pNewNode->key ||
	          ( pNewNode->pLCh->key == pNewNode->key && 
	            pNewNode->pLCh->dim < pNewNode->dim)  );
    swapRCh = pNewNode->pRCh &&
	        ( pNewNode->pRCh->key < pNewNode->key ||
	          ( pNewNode->pRCh->key == pNewNode->key && 
	            pNewNode->pRCh->dim < pNewNode->dim)  );
    if (swapLCh && swapRCh) {
      swapLCh = ( pNewNode->pLCh->key < pNewNode->pRCh->key ||
	          ( pNewNode->pLCh->key == pNewNode->pRCh->key && 
	            pNewNode->pLCh->dim <= pNewNode->pRCh->dim)  );
      swapRCh = !swapLCh;
    } 
  }  /* while */

#ifdef DEBUG
  printf("\n"); 
#endif

  return nSwaps;
}

void hNodeFree(struct hNode *pNode) {
  if (pNode->pRCh) {
    hNodeFree(pNode->pRCh);
  }
  if (pNode->pLCh) {
    hNodeFree(pNode->pLCh);
  }
  myfree(pNode->pY);
  myfree(pNode);
}

void hNodeLFree(struct hNodeL *pNodel) {
  if (pNodel->pRCh) {
    hNodeLFree((struct hNodeL *)pNodel->pRCh);
  }
  if (pNodel->pLCh) {
    hNodeLFree((struct hNodeL *)pNodel->pLCh);
  }
  myfree(pNodel->pY);
  if (pNodel->dim>0) { myfree(pNodel->pA); }
  myfree(pNodel);
}

void heapClear(struct heap *pH) {
  if (pH->pRoot) {
    switch (pH->type) {
      case BASIC:
      case LATTICE:
        hNodeFree(pH->pRoot);
        break;
      case LAST:
        hNodeLFree((struct hNodeL *)pH->pRoot);
        break;
      default:
        break;
    }
    pH->pRoot  = NULL;
    pH->nNodes = 0;
  }
}

void heapDelete(struct heap *pH) {
  if (pH->pRoot) {
    switch (pH->type) {
      case BASIC:
      case LATTICE:
        hNodeFree(pH->pRoot);
        break;
      case LAST:
        hNodeLFree((struct hNodeL *)pH->pRoot);
        break;
      default:
        break;
    }
  }
  myfree(pH);
}

void hNodePrintContents(struct hNode *pNode, int dim, int spdim) {
  int ii, jj;

  if (pNode && 
      (pNode->key || pNode->pLCh || pNode->pY || pNode->dim)) {
    for (ii = 0; ii < dim; ii++) {                  /* tabbing */
      printf("     ");
    }
    printf("%5.2f (%i) [", pNode->key, pNode->dim); /* key and dim */
    for (jj = 0; jj < pNode->dim; jj++) {
      printf(" %5.2f ", pNode->pY[jj]);               /* y */
    }		
    printf("] [");
    for (jj = 0; jj < spdim-pNode->dim; jj++) {
      printf(" %5.2f ", pNode->pZ[jj]);               /* z */
    }		
    printf("]\n");

    if (pNode->pLCh) {
      hNodePrintContents(pNode->pLCh, dim+1, spdim);
    } else {
      printf("\n");
    }
    
    if (pNode->pRCh) {
      hNodePrintContents(pNode->pRCh, dim+1, spdim);
    }

  } else {
    printf("[Empty hNode]\n");
  }
}

void heapPrintContents(struct heap *pH) {
  if (pH) {
    if (pH->nNodes > 0) {
      hNodePrintContents(pH->pRoot, 0, pH->spdim); 
    } else {
      printf("[Empty heap]\n");
    } 
  } else {
    printf("[Null heap]\n");
  }
}

int heapGetSize(struct heap *pH) {
  return pH->nNodes;
}

int heapGetMaxSize(struct heap *pH) {
  return pH->maxNodes;
}

int heapGetDim(struct heap *pH) {
  return pH->spdim;
}

void heapPrintUsage() {
  myprintf("\nUsage:\n\n");
  myprintf("       myHeap = heap('Init', problemDimension [, heapSize], type);\n\n");
  myprintf("       nSwaps = heap('Insert', myHeap,\n");
  myprintf("                     biasedNodeWeight,\n");
  myprintf("                     parentResidualTarget,\n");
  myprintf("                     accumulatedSymbolDecisions,\n");
  myprintf("                     residualDimension\n");
  myprintf("                     [, childOrdinal,\n");
  myprintf("                     unbiasedParentNodeWeight,\n");
  myprintf("                     unconstrainedTarget,\n");
  myprintf("                     previousSiblingDecision\n");
  myprintf("                     [, parentShapingSphereOffset,\n");
  myprintf("                     shapingSphereSquaredRadius,\n");
  myprintf("                     siblingRangeLowerBound,\n");
  myprintf("                     siblingRangeUpperBound,\n");
  myprintf("                     unconstrainedOffset]]);\n\n");
  myprintf("       nSwaps = heap('ReplaceMin', myHeap,\n");
  myprintf("                     bNW, pRT, aSD, rD\n");
  myprintf("                     [, cO, uPNW, uT, pSD\n");
  myprintf("                     [, pSSO, sSSR, sRLB, sRUB, uO]]);\n\n");
  myprintf("       [bNW, pRT, aSD, rD [, cO, uPNW, uT, pSD [, pSSO, sSSR, sRLB, sRUB, uO]]]\n");
  myprintf("              = heap('GetMin', myHeap);\n\n");
  myprintf("       nSwaps = heap('DelMin', myHeap);\n\n");
  myprintf("       nNodes = heap('GetSize', myHeap);\n");
  myprintf("       T      = heap('GetMaxSize', myHeap);\n");
  myprintf("       m      = heap('GetDim', myHeap);\n\n");
  myprintf("                heap('Clear', myHeap);\n");
  myprintf("                heap('Delete', myHeap, 'myHeap');\n");
  myprintf("                heap('Print', myHeap);\n\n");
  myprintf("Notes:\n\n       Input vectors may be columns or rows; output vectors are always columns.\n");
  myprintf("       Operations heap('Insert', ...) and heap('ReplaceMin', ...) return -1 when\n");
  myprintf("       out of memory.  Please send bug reports to karen.su@utoronto.ca.\n\n");
  myerrmsg("Invalid heap operation, please try again.\n\n");
}

#ifdef COMPILEFORMATLAB
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  struct heap   *pH;
  struct hNode  *pNode;
  struct hNodeI *pNodei;
  struct hNodeL *pNodel;
  char           Op[20];
  double         key, *pY, *pZ, pweight, utarget, delta, *pA, sqradius, uoffset;
  int            spdim, T, dim, cnum, ns, nn, lbound, ubound, type, nY, nZ;
  mxArray       *tempArray;

  if (nrhs < 1) { 
    heapPrintUsage(); 
    return;
  }

  /* Get the operation */
  mxGetString(prhs[0],Op,20);

  if (mystrncasecmp(Op,"Init",3) == 0) {
    if (nlhs != 1) {
      type = -1;
    } else if (nrhs == 3) {
      type = (int)mxGetScalar(prhs[2]);
    } else if (nrhs == 4) {
      T    = (int)mxGetScalar(prhs[2]);
      type = (int)mxGetScalar(prhs[3]);
    } else {
      type = -1;
    }

    switch (type) {
      case BASIC:
      case LATTICE:
      case LAST:
        spdim = (int)mxGetScalar(prhs[1]);
        if (nrhs == 3) { pH = heapInit(spdim,type); }
        if (nrhs == 4) { pH = heapInitT(spdim,T,type); }
        break;

      default:
        myprintf("\nUsage: myHeap = heap('Init', problemDimension [, heapSize], type);\n\n");
        myprintf("       Initializes a heap object to handle a decoding problem in pD\n");      
        myprintf("       real dimensions using the OTD algorithm.  Note that myHeap\n");
        myprintf("       will appear in Matlab as an empty matrix.  If the heapSize is\n");      
        myprintf("       not specified, myHeap will have a default maxSize of 32768.\n");      
        myprintf("       Set type == 0 for a lattice decoding problem or type == 1 for\n");      
        myprintf("       a Lattice Space-Time (LAST) decoding problem.\n\n");      
        myerrmsg("Invalid heap operation, please try again.\n\n");
        break;
    }

    plhs[0] = mxCreateDoubleMatrix(0,0,mxREAL);
    mxFree(mxGetPr(plhs[0]));
    mxSetPr(plhs[0],(double *)pH);

  } else if (mystrncasecmp(Op,"Insert",3) == 0) {
    if ((nrhs < 6) || (nlhs != 1)) {
      type = -1;
    } else {
      pH = (struct heap *)mxGetPr(prhs[1]);
      type = pH->type;
      key  = (double)mxGetScalar(prhs[2]);
      pY   = mxGetPr(prhs[3]);
      nY   = mxGetN(prhs[3])*mxGetM(prhs[3]);
      pZ   = mxGetPr(prhs[4]);
      nZ   = mxGetN(prhs[4])*mxGetM(prhs[4]);
      dim  = (int)mxGetScalar(prhs[5]);

      if ((nY != dim) || (nY+nZ != pH->spdim)) {
        myprintf("\nError: Either dim(parentResidualTarget) ~= residualDimension or\n");
        myprintf("       dim(pRT)+dim(accumulatedSymbolDecisions) ~= problemDimension.\n");
        myprintf("       All vectors should be row vectors.\n\n");
        myerrmsg("Invalid heap operation, please try again.\n\n");
      }
    }                                 

    switch (type) {
      case BASIC:
        if (nrhs != 6) {
          myprintf("\nUsage: nSwaps = heap('Insert', myHeap,\n");
          myprintf("                     biasedNodeWeight,\n");
          myprintf("                     parentResidualTarget,\n");
          myprintf("                     accumulatedSymbolDecisions,\n");
          myprintf("                     residualDimension);\n\n");
          myprintf("       Inserts a node into the heap having the specified properties.\n");      
          myprintf("       Returns the number of swaps/bubbleUp operations performed.\n\n");
          myerrmsg("Invalid heap operation, please try again.\n\n");
        }                  

        ns = heapInsertB(key, pY, pZ, dim, pH);
        break;

      case LATTICE:
        if (nrhs != 10) {
          myprintf("\nUsage: nSwaps = heap('Insert', myHeap,\n");
          myprintf("                     biasedNodeWeight,\n");
          myprintf("                     parentResidualTarget,\n");
          myprintf("                     accumulatedSymbolDecisions,\n");
          myprintf("                     residualDimension,\n");
          myprintf("                     childOrdinal,\n");
          myprintf("                     unbiasedParentNodeWeight,\n");
          myprintf("                     unconstrainedTarget,\n");
          myprintf("                     previousSiblingDecision);\n\n");
          myprintf("       Inserts a node into the heap having the specified properties.\n");      
          myprintf("       Returns the number of swaps/bubbleUp operations performed.\n\n");
          myerrmsg("Invalid heap operation, please try again.\n\n");
        }                  

        cnum    = (int)mxGetScalar(prhs[6]);
        pweight = (double)mxGetScalar(prhs[7]);
        utarget = (double)mxGetScalar(prhs[8]);
        delta   = (double)mxGetScalar(prhs[9]);
      
        ns = heapInsertI(key, pY, pZ, dim, cnum, pweight, utarget, delta, pH);
        break;

      case LAST:
        if (nrhs != 15) {
          myprintf("\nUsage: nSwaps = heap('Insert', myHeap,\n");
          myprintf("                     biasedNodeWeight,\n");
          myprintf("                     parentResidualTarget,\n");
          myprintf("                     accumulatedSymbolDecisions,\n");
          myprintf("                     residualDimension,\n");
          myprintf("                     childOrdinal,\n");
          myprintf("                     unbiasedParentNodeWeight,\n");
          myprintf("                     unconstrainedTarget,\n");
          myprintf("                     previousSiblingDecision,\n");
          myprintf("                     parentShapingSphereOffset,\n");
          myprintf("                     shapingSphereSquaredRadius,\n");
          myprintf("                     siblingRangeLowerBound,\n");
          myprintf("                     siblingRangeUpperBound,\n");
          myprintf("                     unconstrainedOffset);\n\n");
          myprintf("       Inserts a node into the heap having the specified properties.\n");      
          myprintf("       Returns the number of swaps/bubbleUp operations performed.\n\n");
          myerrmsg("Invalid heap operation, please try again.\n\n");
        } 

        cnum     = (int)mxGetScalar(prhs[6]);
        pweight  = (double)mxGetScalar(prhs[7]);
        utarget  = (double)mxGetScalar(prhs[8]);
        delta    = (double)mxGetScalar(prhs[9]);
        pA       = mxGetPr(prhs[10]);
        sqradius = (double)mxGetScalar(prhs[11]);
        lbound   = (int)mxGetScalar(prhs[12]);
        ubound   = (int)mxGetScalar(prhs[13]);
        uoffset  = (double)mxGetScalar(prhs[14]);
      
        ns = heapInsertL(key, pY, pZ, dim, cnum, pweight, utarget, delta, pA, sqradius, lbound, ubound, uoffset, pH);
        break;

      default:
        myprintf("\nUsage: nSwaps = heap('Insert', myHeap, ...);\n\n");
        myprintf("       Inserts a node into the heap having the specified properties.\n");      
        myprintf("       Returns the number of swaps/bubbleUp operations performed.\n\n");
        myerrmsg("Invalid heap operation, please try again.\n\n");
        break;
    }

    plhs[0] = mycreatedoublescalar((double)ns);

  } else if (mystrncasecmp(Op,"ReplaceMin",3) == 0) {
    if ((nrhs < 6) || (nlhs != 1)) {
      type = -1;
    } else {
      pH = (struct heap *)mxGetPr(prhs[1]);
      type = pH->type;
      key  = (double)mxGetScalar(prhs[2]);
      pY   = mxGetPr(prhs[3]);
      nY   = mxGetN(prhs[3])*mxGetM(prhs[3]);
      pZ   = mxGetPr(prhs[4]);
      nZ   = mxGetN(prhs[4])*mxGetM(prhs[4]);
      dim  = (int)mxGetScalar(prhs[5]);

      if ((nY != dim) || (nY+nZ != pH->spdim)) {
        myprintf("\nError: Either dim(parentResidualTarget) ~= residualDimension or\n");
        myprintf("       dim(pRT)+dim(accumulatedSymbolDecisions) ~= problemDimension.\n\n");
        myerrmsg("Invalid heap operation, please try again.\n\n");
      }
    }                                

    switch (type) {
      case BASIC:
        if (nrhs != 6) {
          myprintf("\nUsage: nSwaps = heap('ReplaceMin', myHeap,\n");
          myprintf("                     biasedNodeWeight,\n");
          myprintf("                     parentResidualTarget,\n");
          myprintf("                     accumulatedSymbolDecisions,\n");
          myprintf("                     residualDimension);\n\n");
          myprintf("       Replaces the minimum weight (root) node of myHeap by a node\n");      
          myprintf("       having the specified properties.  Returns the number of swaps/\n");      
          myprintf("       bubbleDown operations performed.\n\n");
          myerrmsg("Invalid heap operation, please try again.\n\n");
        }                  

        ns = heapReplaceMinb(key, pY, pZ, dim, pH);
        break;

      case LATTICE:
        if (nrhs != 10) {
          myprintf("\nUsage: nSwaps = heap('ReplaceMin', myHeap,\n");
          myprintf("                     biasedNodeWeight,\n");
          myprintf("                     parentResidualTarget,\n");
          myprintf("                     accumulatedSymbolDecisions,\n");
          myprintf("                     residualDimension,\n");
          myprintf("                     childOrdinal,\n");
          myprintf("                     unbiasedParentNodeWeight,\n");
          myprintf("                     unconstrainedTarget,\n");
          myprintf("                     previousSiblingDecision);\n\n");
          myprintf("       Replaces the minimum weight (root) node of myHeap by a node\n");      
          myprintf("       having the specified properties.  Returns the number of swaps/\n");      
          myprintf("       bubbleDown operations performed.\n\n");
          myerrmsg("Invalid heap operation, please try again.\n\n");
        }                  

        cnum    = (int)mxGetScalar(prhs[6]);
        pweight = (double)mxGetScalar(prhs[7]);
        utarget = (double)mxGetScalar(prhs[8]);
        delta   = (double)mxGetScalar(prhs[9]);

        ns = heapReplaceMini(key, pY, pZ, dim, cnum, pweight, utarget, delta, pH);
        break;

      case LAST:
        if (nrhs != 15) {
          myprintf("\nUsage: nSwaps = heap('ReplaceMin', myHeap,\n");
          myprintf("                     biasedNodeWeight,\n");
          myprintf("                     parentResidualTarget,\n");
          myprintf("                     accumulatedSymbolDecisions,\n");
          myprintf("                     residualDimension,\n");
          myprintf("                     childOrdinal,\n");
          myprintf("                     unbiasedParentNodeWeight,\n");
          myprintf("                     unconstrainedTarget,\n");
          myprintf("                     previousSiblingDecision,\n");
          myprintf("                     parentShapingSphereOffset,\n");
          myprintf("                     shapingSphereSquaredRadius,\n");
          myprintf("                     siblingRangeLowerBound,\n");
          myprintf("                     siblingRangeUpperBound,\n");
          myprintf("                     unconstrainedOffset);\n\n");
          myprintf("       Replaces the minimum weight (root) node of myHeap by a node\n");      
          myprintf("       having the specified properties.  Returns the number of swaps/\n");      
          myprintf("       bubbleDown operations performed.\n\n");
          myerrmsg("Invalid heap operation, please try again.\n\n");
        } 
      
        cnum     = (int)mxGetScalar(prhs[6]);
        pweight  = (double)mxGetScalar(prhs[7]);
        utarget  = (double)mxGetScalar(prhs[8]);
        delta    = (double)mxGetScalar(prhs[9]);
        pA       = mxGetPr(prhs[10]);
        sqradius = (double)mxGetScalar(prhs[11]);
        lbound   = (int)mxGetScalar(prhs[12]);
        ubound   = (int)mxGetScalar(prhs[13]);
        uoffset  = (double)mxGetScalar(prhs[14]);

        ns = heapReplaceMinl(key, pY, pZ, dim, cnum, pweight, utarget, delta, pA, sqradius, lbound, ubound, uoffset, pH);
        break;

      default:
        myprintf("\nUsage: nSwaps = heap('ReplaceMin', myHeap, ...);\n\n");
        myprintf("       Replaces the minimum weight (root) node of myHeap by a node\n");      
        myprintf("       having the specified properties.  Returns the number of swaps/\n");      
        myprintf("       bubbleDown operations performed.\n\n");
        myerrmsg("Invalid heap operation, please try again.\n\n");
        break;
    }

    plhs[0] = mycreatedoublescalar((double)ns);

  } else if (mystrncasecmp(Op,"GetMin",6) == 0) {
    if ((nrhs != 2) || (nlhs < 4)) {
      type = -1;
    } else {
      pH    = (struct heap *)mxGetPr(prhs[1]);
      type  = pH->type;

      pNode = heapGetMin(pH);
      if (pNode) {
        plhs[0] = mycreatedoublescalar((double)(pNode->key));
        plhs[1] = mxCreateDoubleMatrix(pNode->dim,1,mxREAL);
        memcpy(mxGetPr(plhs[1]),pNode->pY,pNode->dim*sizeof(double));
        plhs[2] = mxCreateDoubleMatrix(pH->spdim-pNode->dim,1,mxREAL);
        memcpy(mxGetPr(plhs[2]),pNode->pZ,(pH->spdim-pNode->dim)*sizeof(double));
        plhs[3] = mycreatedoublescalar((double)(pNode->dim));
      } else {
        plhs[0] = mycreatedoublescalar(-1);
        plhs[1] = mxCreateDoubleMatrix(0,0,mxREAL);
        plhs[2] = mxCreateDoubleMatrix(0,0,mxREAL);
        plhs[3] = mxCreateDoubleMatrix(0,0,mxREAL);
      }
    }

    switch (type) {
      case BASIC:
        if (nlhs != 4) {
          myprintf("\nUsage: [biasedNodeWeight,\n");
          myprintf("        parentResidualTarget,\n");
          myprintf("        accumulatedSymbolDecisions,\n");
          myprintf("        residualDimension] = heap('GetMin', myHeap);\n\n");
          myprintf("       Returns the attributes of the node in myHeap having the\n");
          myprintf("       smallest biasedWeight.  The number of swaps/bubbleDown\n");
          myprintf("       operations is also returned.\n\n");
          myerrmsg("Invalid heap operation, please try again.\n\n");
        }
        break;

      case LATTICE:
        if (nlhs != 8) {
          myprintf("\nUsage: [biasedNodeWeight,\n");
          myprintf("        parentResidualTarget,\n");
          myprintf("        accumulatedSymbolDecisions,\n");
          myprintf("        residualDimension,\n");
          myprintf("        childOrdinal,\n");
          myprintf("        unbiasedParentNodeWeight,\n");
          myprintf("        unconstrainedTarget,\n");
          myprintf("        previousSiblingDecision] = heap('GetMin', myHeap);\n\n");
          myprintf("       Returns the attributes of the node in myHeap having the\n");
          myprintf("       smallest biasedWeight.  The number of swaps/bubbleDown\n");
          myprintf("       operations is also returned.\n\n");
          myerrmsg("Invalid heap operation, please try again.\n\n");
        }

        pNodei = (struct hNodeI *)pNode;
        if (pNodei) {
          plhs[4] = mycreatedoublescalar((double)(pNodei->cnum));
          plhs[5] = mycreatedoublescalar((double)(pNodei->pweight));
          plhs[6] = mycreatedoublescalar((double)(pNodei->utarget));
          plhs[7] = mycreatedoublescalar((double)(pNodei->delta));
        } else {
          plhs[4] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[5] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[6] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[7] = mxCreateDoubleMatrix(0,0,mxREAL);
        }
        break;

      case LAST:
        if (nlhs != 13) {
          myprintf("\nUsage: [biasedNodeWeight,\n");
          myprintf("        parentResidualTarget,\n");
          myprintf("        accumulatedSymbolDecisions,\n");
          myprintf("        residualDimension,\n");
          myprintf("        childOrdinal,\n");
          myprintf("        unbiasedParentNodeWeight,\n");
          myprintf("        unconstrainedTarget,\n");
          myprintf("        previousSiblingDecision,\n");
          myprintf("        parentShapingSphereOffset,\n");
          myprintf("        shapingSphereSquaredRadius,\n");
          myprintf("        siblingRangeLowerBound,\n");
          myprintf("        siblingRangeUpperBound,\n");
          myprintf("        unconstrainedOffset] = heap('GetMin', myHeap);\n\n");
          myprintf("       Returns the attributes of the node in the myHeap having the\n");
          myprintf("       smallest biasedWeight.  The number of swaps/bubbleDown\n");
          myprintf("       operations is also returned.\n\n");
          myerrmsg("Invalid heap operation, please try again.\n\n");
        }

        pNodel = (struct hNodeL *)pNode;
        if (pNodel) {
          plhs[4] = mycreatedoublescalar((double)(pNodel->cnum));
          plhs[5] = mycreatedoublescalar((double)(pNodel->pweight));
          plhs[6] = mycreatedoublescalar((double)(pNodel->utarget));
          plhs[7] = mycreatedoublescalar((double)(pNodel->delta));
          plhs[8] = mxCreateDoubleMatrix(pNodel->dim,1,mxREAL);
          memcpy(mxGetPr(plhs[8]),pNodel->pA,pNodel->dim*sizeof(double));
          plhs[9] = mycreatedoublescalar((double)(pNodel->sqradius));
          plhs[10] = mycreatedoublescalar((double)(pNodel->lbound));
          plhs[11] = mycreatedoublescalar((double)(pNodel->ubound));
          plhs[12] = mycreatedoublescalar((double)(pNodel->uoffset));
        } else {
          plhs[4] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[5] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[6] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[7] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[8] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[9] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[10] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[11] = mxCreateDoubleMatrix(0,0,mxREAL);
          plhs[12] = mxCreateDoubleMatrix(0,0,mxREAL);
        }
        break;
        
      default:
        myprintf("\nUsage: [ ... ] = heap('GetMin', myHeap);\n\n");
        myprintf("       Returns the attributes of the node in myHeap having the\n");
        myprintf("       smallest biasedWeight.\n\n");
        myerrmsg("Invalid heap operation, please try again.\n\n");
        break;
    }

  } else if (mystrncasecmp(Op,"DelMin",6) == 0) {
    if ((nrhs != 2) || (nlhs != 1)) {
      type = -1;
    } else {                 
      pH   = (struct heap *)mxGetPr(prhs[1]);
      type = pH->type;
    }

    switch (type) {
      case BASIC:
      case LATTICE:
        ns = heapDelMini(pH);
        break;
      case LAST:
        ns = heapDelMinl(pH);
        break;
      default:
        myprintf("\nUsage: nSwaps = heap('DelMin', myHeap);\n\n");
        myprintf("       Deletes the minimum weight (root) node of myHeap.  Returns the\n");      
        myprintf("       number of swaps/bubbleDown operations performed.\n\n");
        myerrmsg("Invalid heap operation, please try again.\n\n");
        break;
    }

    plhs[0] = mycreatedoublescalar((double)ns);

  } else if (mystrncasecmp(Op,"GetSize",6)==0) {
    if ((nrhs != 2) || (nlhs != 1)) {
      myprintf("\nUsage: nNodes = heap('GetSize', myHeap);\n\n");
      myprintf("       Returns the number of nodes in myHeap.\n\n");
      myerrmsg("Invalid heap operation, please try again.\n\n");
    }

    pH = (struct heap *)mxGetPr(prhs[1]);
    nn = heapGetSize(pH);

    plhs[0] = mycreatedoublescalar((double)nn);

  } else if (mystrncasecmp(Op,"GetMaxSize",6)==0) {
    if ((nrhs != 2) || (nlhs != 1)) {
      myprintf("\nUsage: T = heap('GetMaxSize', myHeap);\n\n");
      myprintf("       Returns the maximum number of nodes stored in myHeap.\n\n");
      myerrmsg("Invalid heap operation, please try again.\n\n");
    }

    pH = (struct heap *)mxGetPr(prhs[1]);
    T  = heapGetMaxSize(pH);

    plhs[0] = mycreatedoublescalar((double)T);

  } else if (mystrncasecmp(Op,"GetDim",6)==0) {
    if ((nrhs != 2) || (nlhs != 1)) {
      myprintf("\nUsage: m = heap('GetDim', myHeap);\n\n");
      myprintf("       Returns the problem dimension associated with myHeap.\n\n");
      myerrmsg("Invalid heap operation, please try again.\n\n");
    }

    pH    = (struct heap *)mxGetPr(prhs[1]);
    spdim = heapGetDim(pH);

    plhs[0] = mycreatedoublescalar((double)spdim);

  } else if (mystrncasecmp(Op,"Clear",3) == 0) {
    if ((nrhs != 2) || (nlhs != 0)) {
      myprintf("\nUsage: heap('Clear', myHeap);\n\n");
      myprintf("       Clears myHeap, i.e., deletes all nodes and frees any memory\n");
      myprintf("       allocated to them.  The heap itself remains in place.\n\n");
      myerrmsg("Invalid heap operation, please try again.\n\n");
    }

    pH = (struct heap *)mxGetPr(prhs[1]);
    pH->pLastDel = NULL;
    heapClear(pH);

  } else if (mystrncasecmp(Op,"Delete",6) == 0) {
    if ((nrhs != 3) || (nlhs != 0)) {
      myprintf("\nUsage: heap('Delete', myHeap, 'myHeap');\n\n");
      myprintf("       Deletes myHeap, i.e., deletes all nodes and frees any memory\n");
      myprintf("       allocated to them, as well as to the heap itself.  The variable\n");
      myprintf("       myHeap is cleared.\n\n");
      myerrmsg("Invalid heap operation, please try again.\n\n");
    }

    pH = (struct heap *)mxGetPr(prhs[1]);
    pH->pLastDel = NULL;
    heapClear(pH); 

    /* tempArray = mxCreateString("H"); */
    /* mxGetName(prhs[1]); */
    mexCallMATLAB(0,NULL,1,(mxArray **)&prhs[2],"clear"); 

  } else if (mystrncasecmp(Op,"Print",3) == 0) {
    if ((nrhs != 2) || (nlhs != 0)) {
      myprintf("\nUsage: heap('Print', myHeap);\n\n");
      myprintf("       Prints some possibly useful information about myHeap.\n\n");
      myerrmsg("Invalid heap operation, please try again.\n\n");
    }

    pH = (struct heap *)mxGetPr(prhs[1]);
    heapPrintContents(pH);

  } else { heapPrintUsage(); }
}
#endif
